import logging

from getpass import getuser
from os import environ
from os.path import dirname, exists

import snackwich
app_path = dirname(snackwich.__file__)

log_paths = [('%s/logs' % (app_path)), 
             '/var/log/snackwich', 
            ]

log_filename = 'snackwich.log'

format = '%(asctime)s  %(levelname)s %(message)s'

for log_path in log_paths:
    if exists(log_path):
        logging.basicConfig(
                level       = logging.DEBUG,
                format      = format,
                filename    = ('%s/%s' % (log_path, log_filename))
            )

        break

# Hook console logging.
#if ('SW_DEBUG' in environ and environ['SW_DEBUG'] or
#    'RI_DEBUG' in environ and environ['RI_DEBUG']
#   ) and 'RI_CONSOLE_LOG_ACTIVE' not in environ:
#    log_console = logging.StreamHandler()
#    log_console.setLevel(logging.DEBUG)
#    log_console.setFormatter(logging.Formatter(format))
#
#    logging.getLogger('').addHandler(log_console)
#
#    environ['RI_CONSOLE_LOG_ACTIVE'] = '1'




"""Adjusted versions of the original functions."""

import types

from snack import ButtonBar, TextboxReflowed, Listbox, GridFormHelp, Grid, \
                  Entry, Label

def ListboxChoiceWindow(screen, title, text, items, buttons = ('Ok', 'Cancel'), 
                        width=40, scroll=0, height=-1, default=None, help=None, 
                        timer_ms=None, secondary_message=None, 
                        secondary_message_width=None):
    """
    - ListboxChoiceWindow(screen, title, text, items, 
            buttons = ('Ok', 'Cancel'), 
            width = 40, scroll = 0, height = -1, default = None,
            help = None):
    """
    
    # Dustin: Added timer_ms parameter. Added secondary_message arguments. 
    #         Added result boolean for whether ESC was pressed. Results are now
    #         dictionaries.
    
    if (height == -1): height = len(items)

    bb = ButtonBar(screen, buttons)
    t = TextboxReflowed(width, text)
    l = Listbox(height, scroll = scroll, returnExit = 1)
    count = 0
    for item in items:
        if (type(item) == types.TupleType):
            (text, key) = item
        else:
            text = item
            key = count

        if (default == count):
            default = key
        elif (default == item):
            default = key

        l.append(text, key)
        count = count + 1

    if (default != None):
        l.setCurrent (default)

    g = GridFormHelp(screen, title, help, 1, 3 + (1 if secondary_message else 
                                                  0))
    
    row = 0
    
    g.add(t, 0, row)
    row += 1

    if secondary_message:
        if not secondary_message_width:
            secondary_message_width = width
    
        t2 = TextboxReflowed(secondary_message_width, secondary_message)
        g.add(t2, 0, row, padding = (0, 0, 0, 1))
        row += 1

    g.add(l, 0, row, padding = (0, 1, 0, 1))
    row += 1
    
    g.add(bb, 0, row, growx = 1)
    row += 1

    if timer_ms:
        g.form.w.settimer(timer_ms)

    rc = g.runOnce()

    return {'button': bb.buttonPressed(rc), 'is_esc': (rc == 'ESC'), 
            'selected': l.current()}

def ButtonChoiceWindow(screen, title, text, buttons=['Ok', 'Cancel'], width=40, 
                       x=None, y=None, help=None, timer_ms=None, 
                       secondary_message=None, secondary_message_width=None):
    """
     - ButtonChoiceWindow(screen, title, text, 
               buttons = [ 'Ok', 'Cancel' ], 
               width = 40, x = None, y = None, help = None):
    """

    # Dustin: Added timer_ms parameter. Added secondary_message arguments. 
    #         Added result boolean for whether ESC was pressed. Results are now
    #         dictionaries.

    bb = ButtonBar(screen, buttons)
    t = TextboxReflowed(width, text, maxHeight = screen.height - 12)

    g = GridFormHelp(screen, title, help, 1, 2 + (1 if secondary_message else 
                                                  0))

    row = 0

    g.add(t, 0, row, padding = (0, 0, 0, 1))
    row += 1

    if secondary_message:
        if not secondary_message_width:
            secondary_message_width = width
    
        t2 = TextboxReflowed(secondary_message_width, secondary_message)
        g.add(t2, 0, row, padding = (0, 0, 0, 1))
        row += 1

    g.add(bb, 0, row, growx = 1)
    row += 1

    if timer_ms:
        g.form.w.settimer(timer_ms)

    rc = g.runOnce(x, y)

    return {'button': bb.buttonPressed(rc), 'is_esc': (rc == 'ESC')}

def EntryWindow(screen, title, text, prompts, allowCancel=1, width=40,
                entryWidth=20, buttons=[ 'Ok', 'Cancel' ], help=None, 
                timer_ms=None, anchorLeft=0, anchorRight=1, 
                secondary_message=None, secondary_message_width=None):

    # Dustin: Added timer_ms parameter. Added secondary_message arguments. 
    #         Added result boolean for whether ESC was pressed. Added 
    #         anchorLeft and anchorRight as arguments to this function. Added 
    #         secondary_message (test below primary text). Results are now
    #         dictionaries.

    bb = ButtonBar(screen, buttons);
    t = TextboxReflowed(width, text)

    count = 0
    for n in prompts:
        count = count + 1

    sg = Grid(2, count)

    count = 0
    entryList = []
    for n in prompts:
        if (type(n) == types.TupleType):
            (n, e) = n
            if (type(e) in types.StringTypes):
                e = Entry(entryWidth, e)
        else:
            e = Entry(entryWidth)

        sg.setField(Label(n), 0, count, padding=(0, 0, 1, 0), 
                    anchorLeft=anchorLeft, anchorRight=anchorRight)
        sg.setField(e, 1, count, anchorLeft = 1)
        count = count + 1
        entryList.append(e)

    g = GridFormHelp(screen, title, help, 1, 3 + (1 if secondary_message else 
                                                  0))
    
    row = 0

    g.add(t, 0, row, padding = (0, 0, 0, 1))
    row += 1

    if secondary_message:
        if not secondary_message_width:
            secondary_message_width = width
    
        t2 = TextboxReflowed(secondary_message_width, secondary_message)
        g.add(t2, 0, row, padding = (0, 0, 0, 1))
        row += 1

    g.add(sg, 0, row, padding = (0, 0, 0, 1))
    row += 1
    
    g.add(bb, 0, row, growx = 1)
    row += 1

    if timer_ms:
        g.form.w.settimer(timer_ms)

    result = g.runOnce()

    entryValues = []
    count = 0
    for n in prompts:
        entryValues.append(entryList[count].value())
        count = count + 1

    return {'button': bb.buttonPressed(result), 'is_esc': (rc == 'ESC'), 
            'entries': tuple(entryValues)
           }





"""A utility function."""

from snack import ButtonBar, TextboxReflowed, Listbox, GridFormHelp, Scale

def ProgressWindow(screen, title, text, progress, max_progress=100, width=40, 
                   help=None, timer_ms=None):
    """
    Render a panel with a progress bar and a "Cancel" button.
    """
    
    scale_proportion = .80
    scale_width = int(width * scale_proportion)
    scale = Scale(scale_width, max_progress)
    scale.set(progress)
    
    bb = ButtonBar(screen, ['Cancel'])
    t = TextboxReflowed(width, text)

    g = GridFormHelp(screen, title, help, 1, 3)
    g.add(t, 0, 0)
    g.add(scale, 0, 1, padding = (0, 1, 0, 1))
    g.add(bb, 0, 2, growx = 1)

    if timer_ms:
        g.form.w.settimer(timer_ms)

    rc = g.runOnce()

    return {'button': bb.buttonPressed(rc), 'is_esc': (rc == 'ESC'), 
            'progress': progress
           }

def MessageWindow(screen, title, text, width=40, help=None, timer_ms=None):
    """
    Render a panel with a message and no buttons. This is intended to
    proceed to the next panel, where some action is taken before 
    returning -its- expression. Meanwhile, this panel is left displayed. 
    Obviously, this panel's timer shouldn't be large, if not zero.
    """
    
    g = GridFormHelp(screen, title, help, 1, 3)

    t = TextboxReflowed(width, text)
    g.add(t, 0, 0)

    if timer_ms:
        g.form.w.settimer(timer_ms)

    rc = g.runOnce()

    return {'is_esc': (rc == 'ESC')}





import logging
import copy

from snack import SnackScreen

from snackwich.patch import ListboxChoiceWindow, ButtonChoiceWindow, \
                            EntryWindow
from snackwich.ui_functions import ProgressWindow, MessageWindow
from snackwich.exceptions import GotoPanelException, \
                                 BreakSuccessionException, \
                                 QuitException, \
                                 RedrawException

class Snackwich(object):
    config = None

    # Which panels follows which (dictionary).
    succession = None

    # The first panel.
    first_key = None

    # Will describe the key of the panel following the current one.
    next_key = None

    # A forward and reverse mapping of expression keys and the expressions' 
    # indexes within the succession.
    keys = None
    keys_r = None

    # Some shorthand aliases to make the config nicer. These must be FROM'd 
    # into the current scope, above.
    aliases = { 'list':     ListboxChoiceWindow,
                'choice':   ButtonChoiceWindow,
                'entry':    EntryWindow,
                'progress': ProgressWindow,
                'message':  MessageWindow,
              }

    def __init__(self, config, forced_succession=None, 
                 forced_first_key=None, config_overlap=None):
        """
        config_filepath: Panel configuration file-path.
        forced_succession: An specific succession of which panels follow which 
                           as an alternative to the automatically-calculated 
                           one. Useful for multiple wizards comprised of a 
                           subset of the windows having a custom order.
        forced_first_key: A specific panel to start with instead of simply the 
                          first one. Again, useful for wizards.
        config_overlap: Dynamic values that will be written on top of the file-
                        loaded values.
        """

        self.succession = forced_succession
        self.first_key = forced_first_key
    
        if config.__class__ != list:
            message = "Config from file-path [%s] is expected to be a list," \
                      "not [%s]." % (config_filepath, \
                                     config.__class__.__name__)

            logging.error(message)
            raise TypeError(message)

        self.config = config

        logging.info("Mapping panel succession.")

        try:
            (self.keys, self.keys_r) = self.__check_succession()
        except:
            logging.exception("Could not calculate panel succession.")
            raise

        if config_overlap:
            for (key, expression) in config_overlap.iteritems():
                if key not in self.keys_r:
                    message = ("Can not overlap panel that does not already "
                               "exist with name [%s]." % (key))

                    logging.error(message)
                    raise Exception(message)

                for (expression_key, expression_value) \
                        in expression.iteritems():
                
                    current_expression = self.config[self.keys_r[key]]
                    current_expression[expression_key] = expression_value

    def __process_stanza(self, screen, key, expression, meta_attributes):
        """Process a single expression (form) from the config."""

        if 'widget' not in meta_attributes:
            message = "'_widget' value is not in the expression for key " \
                      "[%s]." % (key)

            logging.error(message)
            raise Exception(message)
    
        widget = meta_attributes['widget']

        # We provide a couple of aliases in order to be more intuitive. If 
        # used, invoke the actual class, now.
        if isinstance(widget, str):
            if widget not in self.aliases:
                message = "Widget with name [%s] is not a valid alias. " \
                          "Please provide a correct alias or use an actual " \
                          "class." % (widget)
            
                logging.error(message)
                raise Exception(message)
            
            widget = self.aliases[widget]
        
        # Create the widget.
        
        try:
            return widget(screen=screen, **expression)
        except:
            logging.exception("Could not manufacture widget [%s] under key "
                              "[%s]." % (widget.__class__.__name__, key))
            raise

    def __get_current_key(self, expression):
        """For a given expression, validate the structure and return the key.
        """
    
        if isinstance(expression, tuple):
            if len(expression) != 2 or \
               not isinstance(expression[0], str) or \
               not callable(expression[1]):

                message = "As a tuple, an expression must be " \
                          "(<string>, <callable>)."

                logging.error(message)
                raise Exception(message)

            (key, expression_call) = expression
        elif not isinstance(expression, dict):
            message = "Expression is not a callable tuple, nor is it a " \
                      "dictionary."

            logging.error(message)
            raise Exception(message)
        else:
            if '_name' not in expression:
                message = "'_name' key not found in expression."

                logging.error(message)
                raise Exception(message)

            key = expression['_name']

        return key

    def __check_succession(self):
        """Run through the configured panel expressions and determine which 
        leads to which.
        """

        # Produce the succession of keys in a flat list, first (to avoid 
        # calculating each key more than once).

        i = 0
        keys = [ ]
        keys_r = { }
        for expression in self.config:
            try:
                key = self.__get_current_key(expression)
            except:
                logging.exception("Could not render key for expression (%d)." % 
                                  (i))
                raise

            keys.append(key)
            keys_r[key] = i
            i += 1

        if not self.succession:
            num_expressions = len(self.config)
            succession = { }
            i = 0
            while i < num_expressions - 1:
                succession[keys[i]] = keys[i + 1]
                i += 1

            self.succession = succession

        if not self.first_key:
            self.first_key = keys[0] if keys else None

        return (keys, keys_r)

    def __slice_meta_attributes(self, dict_expression):
        meta_attributes = { }
        to_remove = [ ]
        for key in dict_expression:
            if key[0] == '_':
                meta_attributes[key[1:]] = dict_expression[key]
                to_remove.append(key)
        
        for key in to_remove:
            del dict_expression[key]

        return meta_attributes

    def execute(self):
        """Display the panels. This resembles a state-machine, except that the 
        transitions can be determined on the fly.
        """

        results = { }

        # Make sure we actually had panels.
        if not self.first_key:
            return results;

        screen = SnackScreen()

        try:
            # Move through the panels in a linked-list fashion, so that we can 
            # reroute if told to.
        
            key = self.first_key
            quit = False
            while 1:
                logging.info("Processing panel expression with key [%s]." % 
                             (key))

                index = self.keys_r[key]
                expression = copy.deepcopy(self.config[index])

                # Allow there to be a function call or lambda that must be 
                # called to render the actual expression. This call will 
                # receive the past results as well as the current key. This 
                # will allow the current form to feed from previous forms.
                if isinstance(expression, tuple):
                    # The expression is a tuple with the key and a callback.
                
                    (key, expression_call) = expression

                    try:
                        expression = expression_call(self, key, results)
                    except QuitException as e:
                        logging.info("Post-callback for key [%s] has "
                                     "requested emergency exit." % (key))
                        
                        quit = True
                        break
                    except GotoPanelException as e:
                        # Go to a different panel.
                    
                        new_key = e.key
                        
                        logging.info("We were routed from panel with key [%s] "
                                     "to panel with key [%s]." % (key, 
                                                                  new_key))
                        
                        if key not in self.keys_r:
                            message = "We were told to go to panel with " \
                                      "invalid key [%s] while processing " \
                                      "panel with key [%s]." % (new_key, key)
                            
                            logging.error(message)
                            raise Exception(message)
                        
                        key = new_key
                        continue
                    except BreakSuccessionException as e:
                        # Break out of the menu.
                    
                        logging.info("We were told to stop presenting panels "
                                     "while processing panel with key [%s]." % 
                                     (key))
                        break
                    except:
                        logging.exception("Callable expression for key [%s] " 
                                          "threw an exception." % (key))
                        raise

                    if not isinstance(expression, dict):
                        message = "Effective expression for key [%s] is not " \
                                  "a dictionary." % (key)
                        
                        logging.exception(message)
                        raise TypeError(message)
                else:
                    # The expression is a static dictionary.

                    key = expression['_name']
                    del expression['_name']

                try:
                    meta_attributes = self.__slice_meta_attributes(expression)
                except:
                    logging.exception("Could not slice meta-attributes from "
                                      "panel expression with key [%s]." % 
                                      (key))
                    raise

                # Determine the panel to succeed this one. We do this early to 
                # allow callback to adjust this.

                if 'next' in meta_attributes:
                    next_key = meta_attributes['next']
                    
                    if key not in self.keys_r:
                        logging.error("Key [%s] set as next from panel with "
                                      "key [%s] does not refer to a valid "
                                      "panel." % (key))
                    
                    self.next_key = next_key
                else:
                    if key in self.succession:
                        self.next_key = self.succession[key]
                    else:
                        self.next_key = None

                logging.info("Processing expression with key [%s]." % (key))

                try:                
                    result = self.__process_stanza(screen, key, expression, 
                                                   meta_attributes)
# TODO: Move this to a separate call.
                    if 'post_cb' in meta_attributes:
                        callback = meta_attributes['post_cb']
                    
                        if not callable(callback):
                            message = ("Callback for key [%s] is not "
                                       "callable." % (key))
                            
                            logging.error(message)
                            raise Exception(message)

                        try:
                            result_temp = callback(self, key, result, 
                                                   expression)
                            
                            # Allow them to adjust the result, but don't 
                            # require it to be returned.
                            if result_temp:
                                result = result_temp

                        except GotoPanelException as e:
                            # Go to a different panel.
                        
                            new_key = e.key
                            
                            logging.info("We were routed from panel with key "
                                         "[%s] to panel with key [%s] while in"
                                         " post callback." % (key, new_key))
                            
                            if key not in self.keys_r:
                                message = "We were told to go to panel with " \
                                          "invalid key [%s] while in post-" \
                                          "callback for panel with key " \
                                          "[%s]." % (new_key, key)
                                
                                logging.error(message)
                                raise Exception(message)
                            
                            key = new_key
                            continue

                        except RedrawException as e:
                            logging.info("Post-callback for key [%s] has "
                                         "requested a redraw." % (key))
                            
                            # Reset state values that might've been affected by 
                            # a callback.
                            
                            quit = False
                            self.next_key = None
                            
                            continue;

                        except QuitException as e:
                            logging.info("Post-callback for key [%s] has "
                                         "requested emergency exit." % (key))
                            
                            quit = True
                            break
                            
                        except BreakSuccessionException as e:
                            logging.info("Post-callback for key [%s] has "
                                         "instructed us to terminate." % (key))
                            break

                        except:
                            logging.exception("An exception occured in the "
                                              "post-callback for key [%s]." % 
                                              (key))
                            raise

                    # Static expression result handler.
                    else:
                        if 'is_esc' in result and result['is_esc'] or \
                           'button' in result and result['button'] == 'cancel':
                            break

                    if 'collect_results' in meta_attributes and \
                            meta_attributes['collect_results']:
                        if key in results:
                            results[key].append(result)
                        else:
                            results[key] = [result]
                    else:
                        results[key] = result
                            
                except:
                    logging.exception("There was an exception while processing"
                                      " the stanza for key [%s]." % (key))
                    raise

                if self.next_key:
                    key = self.next_key
                else:
                    break
        except:
            logging.exception("There was an exception while processing "
                              "config stanza with key [%s]." % (key))
            raise
        finally:
            screen.finish()

        return None if quit else results

    def set_next_panel(self, key):
        self.next_key = key

